from pathlib import Path
from datetime import datetime
from itertools import chain


class Counter:
    def __init__(self, start=0):
        self.count = start
        self.start = start

    def reset(self):
        self.count = self.start

    def next(self):
        self.count += 1
        return self.count


class Timer:
    def __init__(self):
        self._start_time = None
        self._end_time = None
        self._duration = None

    def reset(self):
        self._start_time = None
        self._end_time = None
        self._duration = None

    def start(self):
        self._start_time = datetime.now()

    def stop(self):
        self._end_time = datetime.now()
        self._duration = self._end_time - self._start_time

    def __str__(self):
        return str(self._duration)


def fetch_file_argument(file_arg, argv=None):
    """
    Arguments for `file_arg` can either be: "0" which indicates to use the first cli arg,
    otherwise, the `file_arg` is expected to be a keyword argument like `--data-file`.

    """
    if argv:
        args = argv
    else:
        import sys

        args = sys.argv[1:]

    err_msg = f"You must provide a data file argument! ({file_arg})."

    if len(args) == 0:
        raise Exception(err_msg.format(file_arg=file_arg))

    first_arg = True if file_arg == "0" else False

    args_normalized = list(chain(*[a.split("=") for a in args]))

    if first_arg:
        file_arg = args[0]

    elif file_arg in args_normalized:
        file_arg = args_normalized[args_normalized.index(file_arg) + 1]

    else:
        raise Exception(err_msg.format(file_arg=file_arg))

    pth = Path(file_arg)

    if pth.exists():
        return pth

    raise Exception(f"Provided path does not exist! {pth}")



import inspect
from pyautogui import FailSafeException
from agw import Automator
from .util import fetch_file_argument

import cutie


class AutomatorApp(object):

    name = "Base Macro Application"

    def __init__(self, *args, ai_class=Automator, cli_args=None):
        self.macros = list(args)
        self.ai_class = ai_class
        self.cli_args = args

    def add_macro(self, macro):
        self.macros.append(macro)

    def __iadd__(self, macro):
        self.macros.append(macro)
        return self

    def display_available_macros(self):
        print("Available Macros:")
        for macro in self.macros:
            #  print(f"  > {macro.name}")
            print(f"  > {macro.__doc__}")
        print("")

    def select_macro(self):
        if len(self.macros) == 1:
            return self.macros[0]
        print("")
        index = cutie.select([n.__doc__ for n in self.macros])
        print("")
        return self.macros[index]

    def run(self):
        try:
            macro = self.select_macro()

            print(f"Running {macro.__doc__} macro...")

            macro_args = inspect.getfullargspec(macro)

            if len(macro_args.args) not in (1, 2):
                raise Exception(f"Macro {macro.__name__} has an invalid signature.")

            ai = self.ai_class()
            ai.request_throttle()

            iterfile_decoration = hasattr(macro, "file_arg")

            if len(macro_args.args) == 2 and iterfile_decoration:
                filepath = fetch_file_argument(macro.file_arg, argv=self.cli_args)
                macro(ai, filepath)
            elif len(macro_args) == 2:
                macro(ai, filepath)
            else:
                macro(ai)

        except KeyboardInterrupt:
            print("Operation cancelled.")

        except FailSafeException:
            print("Fail Safe triggered! Operation cancelled.")

        except Exception as ex:
            print(f"Error: {ex}")



import pyautogui as ag
from time import sleep

__version__ = "0.4.1"


class Cursor(object):
    def to(self, x, y):
        ag.moveTo(x, y)

    def click(self):
        ag.click()

    @property
    def position(self):
        return ag.position()

    def drag_to(self, x, y, speed=0.1, button="left"):
        ag.dragTo(x, y, speed, button=button)

    def drag_rel(self, x, y, speed=0.1, button="left"):
        ag.dragRel(x, y, speed, button=button)


class Keyboard(object):
    def press(self, key, times=1):
        for x in range(0, times):
            ag.press(key)
        return self

    def write(self, text):
        ag.typewrite(text)
        return self

    def writeline(self, text):
        ag.typewrite(text)
        self.enter()
        return self

    def tab(self, times=1):
        self.press("tab")
        return self

    def enter(self, times=1):
        self.press("enter", times)
        return self

    def pagedown(self, times=1):
        self.press("pagedown", times)
        return self

    def pageup(self, times=1):
        self.press("pageup", times)
        return self

    def esc(self, times=1):
        self.press("esc", times)
        return self

    def ctrl(self, key):
        ag.hotkey("ctrl", key)
        return self

    def f(self, number, times=1):
        self.press(f"f{number}", times)
        return self

    def down(self, times=1):
        self.press("down", times)
        return self

    def up(self, times=1):
        self.press("up", times)
        return self

    def right(self, times=1):
        self.press("right", times)
        return self

    def left(self, times=1):
        self.press("left", times)
        return self

    def delete(self, times=1):
        self.press("delete", times)
        return self

    def add(self, times=1):
        self.press("add", times)
        return self

    def subtract(self, times=1):
        self.press("subtract", times)
        return self

    def multiply(self, times=1):
        self.press("multiply", times)
        return self

    def divide(self, times=1):
        self.press("divide", times)
        return self


class MsgBox(object):
    @classmethod
    def confirm(cls, msg):
        return ag.confirm(msg)

    @classmethod
    def alert(cls, msg):
        return ag.alert(msg)

    @classmethod
    def input(cls, msg, title="", default=None):
        return ag.prompt(text=msg, title=title, default=default)


class Automator(object):
    def __init__(self):
        self.cursor = self.cr = Cursor()
        self.keyboard = self.kb = Keyboard()
        self.msgbox = self.mb = MsgBox()
        self.failsafe = True

    @property
    def failsafe(self):
        return ag.FAILSAFE

    @failsafe.setter
    def failsafe(self, value):
        ag.FAILSAFE = value

    @property
    def throttle(self):
        return ag.PAUSE

    @throttle.setter
    def throttle(self, value):
        ag.PAUSE = float(value)

    def request_throttle(self):
        speed = input("  Throttle input (0.1-2.0): ")
        self.throttle = float(speed)

    def pause(self, seconds):
        sleep(seconds)

    def locate_image(self, image_file, region=None):
        return list(ag.locateAllOnScreen(image_file, region=region))

    def image_visible(self, image_file, region=None):
        return len(list(self.locate_image(image_file, region=region)))

    def run(self, macro, *args, **kwargs):
        macro(self, *args, *kwargs)


Automater = Automator



import sys
from pathlib import Path
import cutie
import pyautogui as ag
from .util import ctrl_exit
from .. import screen
from ..util import Timer


def parse_list_of_integers(text):
    current_number = []
    for c in text:
        if c in "0123456789":
            current_number.append(c)
        else:
            if current_number:
                yield int("".join(current_number))
            current_number = []
    if current_number:
        yield int("".join(current_number))


def get_region_input(example_text):
    region = input(f"region {example_text}: ").strip()
    if not region:
        return None
    return list(parse_list_of_integers(region))


def get_filename():
    filename = input("filename: ")
    filename = filename.strip().replace(" ", "_").lower()
    return f"{filename}.png"


def find_image(image_file, region=None):
    timer = Timer()
    timer.start()
    result = screen.locate_image(image_file, region=region)
    timer.stop()
    print(f"   {timer} -> {result}")


def fetch_image_file():
    def list_directory(dirpth):
        dirs = [f for f in dirpth.iterdir() if f.is_dir()]
        files = [
            f
            for f in dirpth.iterdir()
            if f.is_file() and f.suffix.lower() in (".jpg", ".png")
        ]
        dir_pths = [dirpth.parent] + dirs + files

        dir_names = [f"./{f.name}" for f in dirs]
        file_names = [f.name for f in files]
        dir_listing = ["./.."] + dir_names + file_names

        selection = cutie.select(dir_listing, selected_index=0)

        path = dir_pths[selection]

        if path.is_file():
            return path

        return list_directory(path)

    return list_directory(Path.cwd())


def click_location_prior_to_capture():
    if not cutie.prompt_yes_or_no("Click screen location prior to capture?"):
        return

    region = get_region_input("(e.g. 900 500)")
    ag.moveTo(region[0], region[1])
    ag.click()


@ctrl_exit
def capture():
    if not cutie.prompt_yes_or_no("Take screenshot?"):
        print("  > no screenshot taken")
        return

    region = get_region_input("(e.g. 0 0 900 500)")

    filename = get_filename()

    click_spot = click_location_prior_to_capture()

    img = ag.screenshot(region=region)
    img.save(filename)
    print(f"  > screenshot: {filename}")


@ctrl_exit
def find():
    print("Select image file you want to find:")
    image_file = fetch_image_file()
    if not image_file:
        sys.exit()

    search_area = get_region_input("(e.g. 0 0 900 500)")

    click_spot = click_location_prior_to_capture()

    print("Search Whole Screen:")
    find_image(str(image_file))

    print(f"Search Region {search_area}:")
    find_image(str(image_file), region=search_area)



import itertools
from agw import Automater
from .util import ctrl_exit


def move_cursor(cursor, movement):
    def _move_cursor():
        x, y = cursor.position
        return x + movement

    return _move_cursor


@ctrl_exit
def jiggle():
    auto = Automater()
    cursor = auto.cursor

    cycle = itertools.cycle((move_cursor(cursor, 5), move_cursor(cursor, -5)))

    while True:
        x, y = cursor.position
        cursor.to(next(cycle)(), y)
        auto.pause(5)



import sys
from pprint import pprint
from functools import wraps


def ctrl_exit(func):
    @wraps(func)
    def _ctrl_exit(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print("")
            print("\nexiting...\n")
            sys.exit()

    return _ctrl_exit


def output_final_status(kwarg_key, status_msg, object_builder):
    def _output_final_status(func):
        @wraps(func)
        def _output_final_status_(*args, **kwargs):
            kwargs[kwarg_key] = final_obj = object_builder()
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt as ex:
                if final_obj:
                    print("\n{status_msg}:\n")
                    pprint(final_obj)
                    print("*" * 60)
                raise ex

        return _output_final_status_

    return _output_final_status



"""
Usage:
    agw cursor [--prompt-for-name=<int>][--pause=<int>]
    agw jiggle
    agw screenshot (--capture | --find)

Options:
    --prompt-for-name=<int>  Prompt for position name after x seconds. [default: 0]
    --pause=<int>            Length in seconds between recording cursor position. [default: 1]
    --capture                Capture a screen shot.
    --find                   Find position of screen shot image.
    -h --help                Display this screen.
    --version                Display application version.
"""
import docopt
from .cursor import position as cursor
from .jiggler import jiggle
from .screen import capture, find


def main():
    args = docopt.docopt(__doc__)

    if args.get("cursor") is True:
        cursor(
            prompt_for_name=args.get("--prompt-for-name", 0),
            pause_time=args.get("--pause", 1),
        )
    elif args.get("jiggle") is True:
        jiggle()
    elif args.get("screenshot") is True and args.get("--capture"):
        capture()
    elif args.get("screenshot") is True and args.get("--find"):
        find()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['agw', 'agw.cli']

package_data = \
{'': ['*']}

install_requires = \
['Xlib>=0.21,<0.22',
 'cutie>=0.2.2,<0.3.0',
 'docopt>=0.6.2,<0.7.0',
 'pyautogui>=0.9.41,<0.10.0']

extras_require = \
{':sys_platform == "darwin"': ['pyobjc-core>=5.1,<6.0',
                               'pyobjc-framework-Quartz>=5.1,<6.0']}

entry_points = \
{'console_scripts': ['agw = agw.cli:main']}

setup_kwargs = {
    'name': 'agw',
    'version': '0.4.1',
    'description': 'A pyautogui wrapper library for data entry macros.',
    'long_description': None,
    'author': 'Mark Gemmill',
    'author_email': 'contact@markgemmill.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import ctypes
import re


def get_display_name():
    GetUserNameEx = ctypes.windll.secur32.GetUserNameExW
    NameDisplay = 3

    size = ctypes.pointer(ctypes.c_ulong(0))
    GetUserNameEx(NameDisplay, None, size)

    nameBuffer = ctypes.create_unicode_buffer(size.contents.value)
    GetUserNameEx(NameDisplay, nameBuffer, size)
    return nameBuffer.value


EnumWindows = ctypes.windll.user32.EnumWindows

EnumWindowsProc = ctypes.WINFUNCTYPE(
    ctypes.c_bool, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int)
)


def HandleError(func):
    def _handler(*args, **kwargs):
        result = func(*args, **kwargs)
        if result == 0:
            raise Exception("Error!")
        return result

    return _handler


GetWindowText = ctypes.windll.user32.GetWindowTextW
GetWindowTextLength = ctypes.windll.user32.GetWindowTextLengthW
IsWindowVisible = ctypes.windll.user32.IsWindowVisible

CloseWindow = ctypes.windll.user32.CloseWindow
ShowWindow = ctypes.windll.user32.ShowWindow
BringWindowToTop = HandleError(ctypes.windll.user32.BringWindowToTop)
SetForegroundWindow = HandleError(ctypes.windll.user32.SetForegroundWindow)
SetWindowPos = HandleError(ctypes.windll.user32.SetWindowPos)

GetWindowPlacement = ctypes.windll.user32.GetWindowPlacement
GetDesktopWindow = ctypes.windll.user32.GetDesktopWindow
GetClipboardData = ctypes.windll.user32.GetClipboardData


class POINT(ctypes.Structure):
    _fields_ = [("x", ctypes.c_long), ("y", ctypes.c_long)]


class RECT(ctypes.Structure):
    _fields_ = [
        ("left", ctypes.c_long),
        ("top", ctypes.c_long),
        ("right", ctypes.c_long),
        ("bottom", ctypes.c_long),
    ]


class WindowPlacement(ctypes.Structure):
    _fields_ = [
        ("length", ctypes.c_uint),
        ("flags", ctypes.c_uint),
        ("showCmd", ctypes.c_uint),
        ("ptMinPosition", POINT),
        ("ptMaxPosition", POINT),
        ("rcNormalPosition", RECT),
        ("rcDevices", RECT),
    ]

    @property
    def top(self):
        return self.rcNormalPosition.top

    @property
    def left(self):
        return self.rcNormalPosition.left

    @property
    def bottom(self):
        return self.rcNormalPosition.bottom

    @property
    def right(self):
        return self.rcNormalPosition.right

    @staticmethod
    def get(handle):
        winpl = WindowPlacement()
        # print(index, title, str(handle))
        # ShowWindow(handle, 9)
        # BringWindowToTop(handle)
        # SetForegroundWindow(handle)
        GetWindowPlacement(handle, ctypes.byref(winpl))
        return winpl


class Window(object):
    """A wrapper around a Window Handle object.

    """

    def __init__(self, handle):
        self.handle = handle

    @property
    def title(self):
        if not hasattr(self, "_text"):
            length = GetWindowTextLength(self.handle)
            buff = ctypes.create_unicode_buffer(length + 1)
            GetWindowText(self.handle, buff, length + 1)
            self._text = buff.value
        return self._text

    @property
    def placement(self):
        if not hasattr(self, "_placement"):
            self._placement = WindowPlacement.get(self.handle)
        return self._placement

    @property
    def desktop(self):
        if not hasattr(self, "_desktop"):
            self._desktop = Desktop()
        return self._desktop

    @property
    def visible(self):
        return IsWindowVisible(self.handle)

    @property
    def top(self):
        return self.placement.top

    @property
    def right(self):
        return self.placement.right

    @property
    def bottom(self):
        return self.placement.bottom

    @property
    def left(self):
        return self.placement.left

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    def bring_to_front(self):
        ShowWindow(self.handle, 9)
        SetForegroundWindow(self.handle)

    def set_position(self, top, left, width, height):
        SetWindowPos(self.handle, 0, top, left, width, height, 0)

    def anchor_top_left(self):
        width = int(self.desktop.width / 2)
        height = int(self.desktop.height / 2)
        self.set_position(0, 0, width, height)

    def anchor_left(self):
        width = int(self.desktop.width / 2)
        self.set_position(0, 0, width, self.desktop.height)

    def get_window_text(handle):
        length = GetWindowTextLength(handle)
        buff = ctypes.create_unicode_buffer(length + 1)
        GetWindowText(handle, buff, length + 1)
        return buff.value

    @staticmethod
    def iter_windows():
        windows = []

        def enumerate_windows(hwnd, lParam):
            windows.append(Window(hwnd))
            return True

        EnumWindows(EnumWindowsProc(enumerate_windows), 0)

        for entry in windows:
            yield entry

    @staticmethod
    def find(name):
        windows = list(Window.find_all(name))
        if windows:
            return windows[0]
        return []

    @staticmethod
    def find_all(name):
        def _match_in(title):
            return name in title

        matcher = _match_in

        if isinstance(name, re.Pattern):
            regex = re.compile(name)

            def _match_re(title):
                return regex.match(title)

            matcher = _match_re

        for wind in Window.iter_windows():
            if not wind.visible:
                continue
            if matcher(wind.title):
                yield wind


class Desktop(Window):
    def __init__(self):
        dsktp = GetDesktopWindow()
        super(Desktop, self).__init__(dsktp)


if __name__ == "__main__":

    for win in Window.iter_windows():
        if win.visible:
            print(win.title)

    chrome = Window.find(re.compile("^.* - Google Chrome$"))

    print(f"{chrome.title}")
    print(f"Chrome: {chrome.top} {chrome.left}, {chrome.width} x {chrome.height}")

    chrome.bring_to_front()
    chrome.anchor_left()



import pyautogui as ag
from pyscreeze import ImageNotFoundException


class Region:

    def __init__(self, x, y, x1=None, y1=None, w=None, h=None):
        self.x = y
        self.y = y

        if x1 and y1 and not w and not h:
            self.x1 = x1
            self.y1 = y1
            self.w = x1 - x
            self.h = y1 - y
        elif w and h and not x1 and not y1:
            self.w = w
            self.h = h
            self.x1 = x + w
            self.y1 = y + h
        else:
            raise Exception("Either supply x1 and y1 coordinates, or the width and high values. Not both")

    @property
    def box(self):
        return (self.x, self.y, self.w, self.h)

    @staticmethod
    def from_coordinates(x, y, x1, y1):
        return Region(x, y, x1=x1, y1=y1)

    @staticmethod
    def from_measurements(x, y, w, h):
        return Region(x, y, w=w, h=h)



def locate_image(image, region=None):
    try:
        if isinstance(region, Region):
            region = region.box
        return ag.locateOnScreen(image, region=region)
    except ImageNotFoundException:
        return None


def image_located(image, region=None, note=None, action=None):

    result = locate_image(image, region=region)

    if result and note:
        print(note)

    if result and action:
        action()

    return result



import sys
import cutie
from collections import namedtuple
from functools import wraps
from pathlib import Path
from agw import Automater
from .util import Counter


def data_generator(fields, header_row=True, separator=","):

    DataObject = namedtuple("DataObject", fields)

    def iter_lines(filepath):
        with filepath.open(mode="r") as fh_:
            for index, line in enumerate(fh_.readlines()):
                if index == 0 and header_row is True:
                    continue
                if not line or line.startswith("#"):
                    continue
                data = DataObject(*[i.strip() for i in line.strip().split(separator)])
                yield index, data

    return iter_lines


class MacroDataFileHandler(object):
    def __init__(
        self,
        data_fields,
        header_row=False,
        separator=",",
        line_pause=False,
        line_start=False,
    ):
        self.data_fields = data_fields
        self._iterate = data_generator(data_fields, header_row, separator)
        self.line_pause = 0
        self.line_start = 0

    def get_inputs(self):
        line_pause = cutie.get_number("  Pause after every N lines: ")
        self.line_pause = int(line_pause) if line_pause else None

        skip_to_line = cutie.get_number("  Start from line: ")
        self.skip_to_line = int(skip_to_line) if skip_to_line else None

    def iter(self, filepath):
        filepath = Path(filepath)
        ai = Automater()

        line_count = Counter()
        pause_count = Counter()

        for index, data in self._iterate(filepath):

            if self.skip_to_line and line_count.next() < self.skip_to_line:
                continue

            if self.line_pause and pause_count.next() == self.line_pause:
                pause_count.reset()
                if ai.msgbox.confirm("continue?") == "No":
                    print("Operation concelled.")
                    sys.exit(0)

            yield index, data


def iterfile(data_fields, file_arg="--data-file", header_row=False, separator=","):
    """Convenience decorator that wrapps creation of file iterator.
    """

    def _iterfile(func):

        handler = MacroDataFileHandler(
            data_fields, header_row=header_row, separator=separator
        )

        @wraps(func)
        def __iterfile(ai, filepath):
            iterator = handler.iter(filepath)
            ai.get_data_handling_inputs = handler.get_inputs
            return func(ai, iterator)

        __iterfile.file_arg = file_arg

        return __iterfile

    return _iterfile



import sys
from time import sleep
from agw import Cursor
from agw import MsgBox
from .util import ctrl_exit
from .util import output_final_status


@ctrl_exit
@output_final_status("named_positions", "Named Positions", dict)
def position(*args, **kwargs):

    cursor = Cursor()
    last_position = (-1, -1)

    prompt_after_seconds = int(kwargs.get("prompt_for_name", 0))
    pause_time = float(kwargs.get("pause_time", 1))
    captured_named_positions = kwargs["named_positions"]

    prompt_count = 0

    while True:

        current_position = cursor.position

        if current_position == last_position:
            sys.stdout.write("*")
            sys.stdout.flush()
            prompt_count += 1

            if prompt_after_seconds > 0 and prompt_count == prompt_after_seconds:
                position_name = MsgBox.input(
                    f"Label current position ({current_position[0]}, {current_position[1]}):"
                )

                if position_name:
                    captured_named_positions[position_name] = (
                        current_position[0],
                        current_position[1],
                    )
                    print(
                        f"\n{position_name}: {current_position[0]} {current_position[1]}"
                    )
                prompt_count = 0
                last_position = (-1, -1)

        else:
            last_position = current_position
            x, y = current_position
            sys.stdout.write("\n")
            sys.stdout.write(f"x: {x: >4}  y: {y: >4} ")
            sys.stdout.flush()

        sleep(pause_time)
